#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/client.py $
$Id: client.py 26303 2005-03-08 14:08:46Z dbinger $
"""
import sys
import os
from code import InteractiveConsole
from durus.utils import p64, u64
from durus.file_storage import FileStorage
from durus.client_storage import ClientStorage
from durus.connection import Connection
from durus.storage_server import DEFAULT_PORT, DEFAULT_HOST, wait_for_server
from pprint import pprint

def configure_readline(namespace, history_path):
    try:
        import readline, rlcompleter, atexit
        readline.set_completer(
            rlcompleter.Completer(namespace=namespace).complete)
        readline.parse_and_bind("tab: complete")
        def save_history(history_path=history_path):
            readline.write_history_file(history_path)
        atexit.register(save_history)
        if os.path.exists(history_path):
            readline.read_history_file(history_path)
    except ImportError:
        pass

def interactive_client(file, host, port, cache_size, readonly, repair,
                       startup):
    if file:
        storage = FileStorage(file, readonly=readonly, repair=repair)
        description = file
    else:
        wait_for_server(host, port)
        storage = ClientStorage(host=host, port=port)
        description = "%s:%s" % (host, port)
    connection = Connection(storage, cache_size=cache_size)
    namespace = {'connection': connection,
                 'root': connection.get(0),
                 'get': connection.get,
                 'sys': sys,
                 'os': os,
                 'p64': p64,
                 'u64': u64,
                 'pp': pprint}
    configure_readline(namespace, os.path.expanduser("~/.durushistory"))
    console = InteractiveConsole(namespace)
    if startup:
        console.runsource('execfile("%s")' % os.path.expanduser(startup))
    help = ('    connection -> the connection\n'
            '    root       -> get(0)\n'
            '    get(oid)   -> get an object\n'
            '    pp(object) -> pretty-print')
    console.interact('Durus (%s)\n%s' % (description, help))

def client_main():
    from optparse import OptionParser
    parser = OptionParser()
    parser.set_description("Opens a client connection to a Durus server.")
    parser.add_option(
        '--file', dest="file", default=None,
        help="If this is not given, the storage is through a Durus server.")
    parser.add_option(
        '--port', dest="port", default=DEFAULT_PORT,
        type="int",
        help="Port the server is on. (default=%s)" % DEFAULT_PORT)
    parser.add_option(
        '--host', dest="host", default=DEFAULT_HOST,
        help="Host of the server. (default=%s)" % DEFAULT_HOST)
    parser.add_option(
        '--cache_size', dest="cache_size", default=10000,
        type="int",
        help="Host of the server. (default=0, no limit)")
    parser.add_option(
        '--repair', dest='repair', action='store_true',
        help=('Repair the filestorage by truncating to remove anything '
              'that is malformed.  Without this option, errors '
              'will cause the program to report and terminate without '
              'attempting any repair.'))
    parser.add_option(
        '--readonly', dest='readonly', action='store_true',
        help='Open the file in read-only mode.')
    parser.add_option(
        '--startup', dest='startup',
        default=os.environ.get('DURUSSTARTUP', ''),
        help=('Full path to a python startup file to execute on startup.'
              '(default=DURUSSTARTUP from environment, if set)')
        )
    (options, args) = parser.parse_args()
    interactive_client(options.file, options.host, options.port,
                       options.cache_size, options.readonly, options.repair,
                       options.startup)


if __name__ == '__main__':
    client_main()



"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/persistent.py $
$Id: persistent.py 24882 2004-08-10 14:52:48Z dbinger $
"""

from durus.utils import format_oid

# these must match the constants in _persistent.c
UNSAVED = 1
SAVED = 0
GHOST = -1

try:
    from _persistent import PersistentBase
except ImportError:
    _GHOST_SAFE_ATTRIBUTES = {
        '__repr__': 1,
        '__class__': 1,
        '__setstate__': 1,
    }

    class PersistentBase(object):
        """
        The faster implementation of this class is in _persistent.c.
        The __slots__ and methods of this class are the ones that typical
        applications use very frequently, so we want them to be fast.

        Instance attributes:
          _p_status: UNSAVED | SAVED | GHOST
            UNSAVED means that state that is here, self.__dict__, is usable and
              has not been stored.
            SAVED means that the state that is here, self.__dict__, is usable
              and the same as the stored state.
            GHOST means that the state that is here, self.__dict__, is empty
              and unusable until it is updated from storage.

            New instances are UNSAVED.
            UNSAVED -> SAVED
              happens when the instance state, self.__dict__, is saved.
            UNSAVED -> GHOST
              happens on an abort (if the object has previously been saved).
            SAVED   -> UNSAVED
              happens when changes are made to self.__dict__.
            SAVED   -> GHOST
              happens when the cache manager wants space.
            GHOST   -> SAVED
              happens when the instance state is loaded from the storage.
            GHOST   -> UNSAVED
              this happens when you want to make changes to self.__dict__.
              The stored state is loaded during this state transition.
            
          _p_touched: 0 | 1
            _p_touched is set to 1 on every access to an attribute in
            self.__dict__.  This allows the cache manager to identify SAVED
            instances that have not been recently used.
        """

        __slots__ = ['_p_status', '_p_touched']

        def __getattribute__(self, name):
            if name[:3] != '_p_' and name not in _GHOST_SAFE_ATTRIBUTES:
                if self._p_status == GHOST:
                    self._p_load_state()
                self._p_touched = 1
            return object.__getattribute__(self, name)

        def __setattr__(self, name, value):
            if name[:3] != '_p_' and name not in _GHOST_SAFE_ATTRIBUTES:
                self._p_note_change()
            object.__setattr__(self, name, value)



class Persistent(PersistentBase):
    """
    All Durus persistent objects should inherit from this class.

    Instance attributes:
      _p_connection: durus.connection.Connection | None
        The Connection to the Storage that stores this instance.
        The _p_connection is None when this instance has never been stored.
      _p_oid: str | None
        The identifier assigned when the instance was first stored.
        The _p_oid is None when this instance has never been stored.
    """

    __slots__ = ['_p_connection',
                 '_p_oid',
                 '__dict__',    # Provides for a normal instance __dict__.
                 '__weakref__'] # The cache uses weak references.

    def __new__(klass, *args, **kwargs):
        instance = PersistentBase.__new__(klass, *args, **kwargs)
        instance._p_status = UNSAVED
        instance._p_touched = 1
        instance._p_connection = None
        instance._p_oid = None
        return instance

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        object.__getattribute__(self, '__dict__').update(state)

    # This is here for ZODB-compatibility.
    # Note that setting _p_changed to a non-true value does nothing.
    _p_changed = property(
        lambda self: self._p_status == UNSAVED,
        lambda self, value: value and self._p_note_change())

    def __repr__(self):
        if self._p_oid is None:
            identifier = '@%x' % id(self)
        else:
            identifier = self._p_format_oid()
        return "<%s %s>" % (self.__class__.__name__, identifier)

    def __delattr__(self, name):
        self._p_note_change()
        PersistentBase.__delattr__(self, name)

    def _p_load_state(self):
        assert self._p_status == GHOST
        self._p_connection.load_state(self)
        self._p_set_status_saved()

    def _p_note_change(self):
        if self._p_status != UNSAVED:
            if self._p_status == GHOST:
                self._p_load_state()
            self._p_connection.note_change(self)
            self._p_status = UNSAVED

    def _p_format_oid(self):
        return format_oid(self._p_oid)

    def _p_set_status_ghost(self, getattribute=object.__getattribute__):
        d = getattribute(self, '__dict__')
        d.clear()
        self._p_status = GHOST

    def _p_set_status_saved(self):
        self._p_connection.note_saved(self)
        self._p_status = SAVED

    def _p_set_status_unsaved(self):
        if self._p_status == GHOST:
            self._p_load_state()
        self._p_status = UNSAVED

    def _p_is_ghost(self):
        return self._p_status == GHOST

    def _p_is_unsaved(self):
        return self._p_status == UNSAVED

    def _p_is_saved(self):
        return self._p_status == SAVED


_Marker = object()

class ComputedAttribute(Persistent):
    """Computed attributes do not have any state that needs to be saved in
    the database.  Instead, their value is computed based on other persistent
    objects.  Although they have no real persistent state, they do have
    OIDs.  That allows synchronize of the cached value between connections
    (necessary to maintain consistency).  If the value becomes invalid in one
    connection then it must be invalidated in all connections.  That is
    achieved by marking the object as UNSAVED and treating it like a normal
    persistent object.

    Instance attributes: none
    """

    def __getstate__(self):
        return None

    def __setstate__(self, state):
        assert state is None

    def _p_load_state(self):
        # don't need to read state from connection, there is none
        self._p_set_status_saved()

    def invalidate(self):
        """Forget value and mark object as UNSAVED.  On commit it will cause
        other connections to receive a invalidation notification and forget the
        value as well.
        """
        self.__dict__.clear()
        self._p_note_change()

    def get(self, compute):
        """(compute) -> value

        Compute the value (if necessary) and return it.  'compute' needs
        to be a function that takes no arguments.
        """
        # we are careful here not to mark object as UNSAVED
        d = self.__dict__
        value = d.get('value', _Marker)
        if value is _Marker:
            value = d['value'] = compute()
        return value



#!/usr/bin/env python
"""
This script is for converting an existing durus database to the
other durus FileStorage format.
If the existing database is in FileStorage1 format, it is
converted to FileStorage2 format.
If the existing database is in FileStorage2 format, it is
converted to FileStorage1 format.
All objects are put in one new transaction with a zero tid,
so this removes obsolete object records.
"""
from durus.connection import Connection
from datetime import datetime
from durus.file_storage import FileStorage, FileStorage1, FileStorage2
from optparse import OptionParser
from os.path import exists
from os import rename

def duplicate_file_storage(old_storage, new_storage):
    """(old_storage:FileStorage, new_storage:FileStorage)
    Transfer the records from a storage to a new one.
    """
    assert len(new_storage.index) == 0, "new_storage must be empty"
    # Transfer the records
    new_storage.begin()
    for oid, record in old_storage.gen_oid_record():
        new_storage.store(oid, record)
    new_storage.end()

def repickle_storage(storage):
    """(storage: FileStorage)
    Force very object to be loaded and re-pickled.
    This also packs, so that all of the old pickles are removed.
    """
    connection = Connection(storage)
    for j, oid in enumerate(storage.index):
        obj = connection.get(oid)
        obj._p_note_change()
        if j and j % 10000 == 0:
            connection.commit()
    connection.commit()
    storage.pack()

def move_to_backup(name):
    if exists(name):
        rename(name, "%s.%s" % (name, datetime.now()))

def copy_to_new_format(from_file, to_file, format):
    tmp_file_name = "%s.%s.tmp" % (to_file, datetime.now())
    if format == 1:
        to_storage = FileStorage1(tmp_file_name)
    elif format == 2:
        to_storage = FileStorage2(tmp_file_name)
    from_storage = FileStorage(from_file, readonly=True)

    duplicate_file_storage(from_storage, to_storage)
    old_num_records = len(from_storage.index)
    assert len(to_storage.index) == old_num_records
    from_storage.close()
    repickle_storage(to_storage)
    assert len(to_storage.index) > max(1, old_num_records/2)
    move_to_backup(to_file)
    rename(tmp_file_name, to_file)

if __name__ == '__main__':
    parser = OptionParser()
    parser.set_description("Copies a file storage using a specified format.")
    parser.add_option(
        '--from', dest="from_file", default=None,
        help="The durus storage file to be copied. (required)")
    parser.add_option(
        '--to', dest="to_file", default=None,
        help="The target file. (required)")
    parser.add_option(
        '--format', default=2,
        type=int,
        help="The FileStorage format to use (1 or 2).  The default is 2.")
    (options, args) = parser.parse_args()
    if None in (options.from_file, options.to_file):
        print "\nThe --from and --to arguments are required.\n"
        print parser.format_help()
        raise SystemExit
    copy_to_new_format(options.from_file, options.to_file, options.format)




"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/utils.py $
$Id: utils.py 23882 2004-04-05 20:47:59Z dbinger $
"""

import struct

def format_oid(oid):
    """(oid:str) -> str
    Returns a nice representation of an 8-byte oid.
    """
    return str(oid and u64(oid))

def p64(v):
    """Pack an integer or long into a 8-byte string"""
    return struct.pack(">Q", v)

def u64(v):
    """Unpack an 8-byte string into a 64-bit long integer."""
    return struct.unpack(">Q", v)[0]

def p32(v):
    """Pack an integer or long into a 4-byte string"""
    return struct.pack(">L", v)

def u32(v):
    """Unpack an 8-byte string into a 32-bit long integer."""
    return struct.unpack(">L", v)[0]




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/test/utest_file_storage.py $
$Id: utest_file_storage.py 27100 2005-07-29 20:54:23Z dbinger $
"""
from sancho.utest import UTest
from durus.file_storage import TempFileStorage, FileStorage
from durus.file_storage import FileStorage1
from durus.serialize import pack_record
from durus.utils import p64


class Test (UTest):

    def check_file_storage(self):
        self._check_file_storage(TempFileStorage())
        self._check_file_storage(FileStorage1())

    def _check_file_storage(self, storage):
        b = storage
        assert b.new_oid() == p64(1)
        assert b.new_oid() == p64(2)
        try:
            b.load(p64(0))
            assert 0
        except KeyError: pass
        record = pack_record(p64(0), 'ok', '')
        b.store(p64(0), record)
        b.begin()
        b.end()
        b.sync()
        b.begin()
        b.store(p64(1), pack_record(p64(1), 'no', ''))
        b.end()
        assert len(list(b.gen_oid_record())) == 2
        b.pack()
        import durus.file_storage
        if durus.file_storage.RENAME_OPEN_FILE:
            durus.file_storage.RENAME_OPEN_FILE = False
            b.pack()
            c = FileStorage(b.get_filename(), readonly=True)
            try:
                c.pack()
                assert 0 
            except IOError: # read-only storage
                pass
        b.close()
        try:
            b.pack()
            assert 0 
        except IOError: # storage closed
            pass
        try:
            b.load(0)
            assert 0
        except IOError: # storage closed
            pass


if __name__ == "__main__":
    Test()




"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/logger.py $
$Id: logger.py 24062 2004-04-27 19:18:06Z dbinger $
"""

import sys
from logging import getLogger, StreamHandler, Formatter, INFO

logger = getLogger('durus')
log = logger.log

def direct_output(file):
    logger.handlers[:] = []
    handler = StreamHandler(file)
    handler.setFormatter(Formatter("%(message)s"))
    logger.addHandler(handler)
    logger.propagate = False
    logger.setLevel(INFO)
    if file is sys.__stderr__:
        return
    if sys.stdout is sys.__stdout__:
        sys.stdout = file
    else:
        log(100, "sys.stdout already customized.")
    if sys.stderr is sys.__stderr__:
        sys.stderr = file
    else:
        log(100, "sys.stderr already customized.")

if not logger.handlers:
    direct_output(sys.stderr)

def is_logging(level):
    return logger.getEffectiveLevel() <= level




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/test/utest_storage.py $
$Id: utest_storage.py 27487 2005-09-29 13:18:29Z dbinger $
"""
from sancho.utest import UTest, raises
from durus.storage import MemoryStorage
from durus.serialize import pack_record
from durus.utils import p64

class Test (UTest):

    def check_memory_storage(self):
        b = MemoryStorage()
        assert b.new_oid() == p64(1)
        assert b.new_oid() == p64(2)
        raises(KeyError, b.load, p64(0))
        record = pack_record(p64(0), 'ok', '')
        b.begin()
        b.store(p64(0), record)
        b.end()
        b.sync()
        b.begin()
        b.store(p64(1), pack_record(p64(1), 'no', ''))
        b.end()
        assert len(list(b.gen_oid_record())) == 2
        assert record == b.load(p64(0))

if __name__ == "__main__":
    Test()




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/test/utest_persistent.py $
$Id: utest_persistent.py 27079 2005-07-25 20:54:05Z dbinger $
"""
from cPickle import dumps, loads
from durus.connection import Connection
from durus.file_storage import TempFileStorage
from durus.persistent import Persistent
from durus.utils import p64
from sancho.utest import UTest


class Test (UTest):

    def check_getstate(self):
        p=Persistent()
        assert p.__getstate__() == {}
        p.a = 1
        assert p.__getstate__() == {'a':1}

    def check_setstate(self):
        p=Persistent()
        p.__setstate__({})
        p.__setstate__({'a':1})
        assert p.a == 1

    def check_change(self):
        p=Persistent()
        p._p_changed == 0
        p._p_note_change()
        assert p._p_changed == True

    def check_accessors(self):
        p=Persistent()
        p._p_oid
        assert p._p_format_oid() == 'None'
        p._p_oid = p64(1)
        assert p._p_format_oid() == '1'
        assert repr(p) == "<Persistent 1>"

    def check_more(self):
        storage = TempFileStorage()
        connection = Connection(storage)
        root=connection.get_root()
        assert root._p_is_ghost()
        root.a = 1
        assert root._p_is_unsaved()
        del root.a
        connection.abort()
        assert root._p_is_ghost()
        try:
            root.a
            assert 0
        except AttributeError: pass
        root._p_set_status_saved()
        assert root._p_is_saved()
        root._p_set_status_unsaved()
        assert root._p_is_unsaved()
        root._p_set_status_ghost()
        assert root._p_is_ghost()
        root._p_set_status_unsaved()

    def pickling(self):
        a = Persistent()
        pickle_a = dumps(a, 2) # Pickle protocol 2 is required.
        b = loads(pickle_a)
        assert isinstance(b, Persistent)

if __name__ == "__main__":
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/test/utest_client_storage.py $
$Id: utest_client_storage.py 27079 2005-07-25 20:54:05Z dbinger $
"""
from durus import run_durus
from durus.client_storage import ClientStorage
from durus.error import ReadConflictError
from durus.serialize import pack_record
from durus.utils import p64
from popen2 import Popen4
from sancho.utest import UTest, raises
from sets import Set
from time import sleep

class Test (UTest):

    def _pre(self):
        self.port = 9123
        self.server = Popen4('python %s --port=%s' % (
            run_durus.__file__, self.port))
        sleep(3) # wait for bind

    def _post(self):
        run_durus.stop_durus("", self.port)

    def check_client_storage(self):
        b = ClientStorage(port=self.port)
        c = ClientStorage(port=self.port)
        assert b.new_oid() == p64(1)
        assert b.new_oid() == p64(2)
        try:
            b.load(p64(0))
            assert 0
        except KeyError: pass
        record = pack_record(p64(0), 'ok', '')
        b.begin()
        b.store(p64(0), record)
        assert b.end() is None
        b.load(p64(0))
        assert b.sync() == []
        b.begin()
        b.store(p64(1), pack_record(p64(1), 'no', ''))
        b.end()
        assert len(list(b.gen_oid_record())) == 2
        b.pack()
        assert len(list(b.gen_oid_record())) == 1
        raises(ReadConflictError, c.load, p64(0))
        raises(ReadConflictError, c.load, p64(0))
        assert Set(c.sync()) == Set([p64(0), p64(1)])
        assert record == c.load(p64(0))

if __name__ == "__main__":
    Test()

